module m_erosion

contains
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              ||===\\                                                         |
!              ||    \\                                                        |
!              ||     ||   //==\\   ||  ||   //==||  ||/==\\                   |
!              ||     ||  ||    ||  ||  ||  ||   ||  ||    ||                  |
!              ||    //   ||    ||  ||  ||  ||   ||  ||                        |
!              ||===//     \\==//    \\==\\  \\==\\  ||                        |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              EROSION    Nov. 2006                                            |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

subroutine erosion (surface, olsf, is, params, converge)

!------------------------------------------------------------------------------|
!(((((((((((((((( Purpose of the routine  ))))))))))))))))))))))))))))))))))))))
!------------------------------------------------------------------------------|
! Erosion has two modes: Flat Erosion (everything above zerosion shaved off) and
! erosion by Cascade.

! surface are the surface/sheet object
! (the top free surface)
! olsf is a octreelsf object containig the geometry of the current velocity octree/object
! is is the number of the surface (0 playing a special role)
! params%zerosion is the level of erosion (between 0 and 1)

! Note that all surfaces are required because they can be affected by the erosion process

! for the top surface, we apply the erosion and on the octree
! we define an lsf that is simply the height of the eroded surface
! This allows us, for the other surfaces to find the height of the
! free/eroded surface at any point in space and check whether the
! vertical position of the point is above or below the free surface
! We then simply "erode" the following surfaces using this information

! flat_erosion is the parameter to switch beween the two modes.
! length_scale is the length scale in km (corresponding to the unit length of the box)
! velocity_scale is the scale for the velocity in km/Myr
! dt is the length of the dimensionless time step
! fluvial_erosion is the fluvial erosion constant
! diffusion_erosion is the diffusion erosion constant

!------------------------------------------------------------------------------|
!((((((((((((((((  declaration of the subroutine arguments  ))))))))))))))))))))
!------------------------------------------------------------------------------|

use definitions
use m_toolbox
use m_octree
use m_fastscape_solo
use m_define_erosion_parameter
!use mpi

implicit none

include 'mpif.h'

type (sheet) surface
type (octreelsf) olsf
type (parameters) params
integer is
logical :: converge

!------------------------------------------------------------------------------|
!(((((((((((((((( declaration of the subroutine internal variables )))))))))))))
!------------------------------------------------------------------------------|
integer ierr,iproc,nproc,i,nt,nh,nohalt_hull,loc,nnn,nnlist,ntrilist,k,kp,ic,icp
integer ntmax,nhmax,npmax,nmax,nmode,nvmax,nnpnmax,int_method
double precision,dimension(:),allocatable::field,wpoints
double precision,dimension(:,:),allocatable::points,centres,pointsmin,pointsmax,ppoints
integer,dimension(:,:),allocatable::vertices,neighbour
integer,dimension(:),allocatable::hulltriangles,ln_work
logical*1,dimension(:),allocatable::lt_work_l,ln_work_l
integer,dimension(:),allocatable::vis_tlist,vis_elist,add_tlist
logical clockwise,out,nnext
double precision eps
integer dmode,err
double precision x,y,f,df(2),ddf(3),v
double precision,dimension(:,:),allocatable::work_d1,work_d2,work_d3
double precision,dimension(:,:),allocatable::work_d4,work_d5,work_d6,work_d7
integer,dimension(:),allocatable::work_i1,work_i3
integer,dimension(:,:),allocatable::work_i2
real,dimension(:),allocatable::work_r2
real,dimension(:,:),allocatable::work_r1
double precision,dimension(:),allocatable::lsf
double precision,dimension(:),allocatable::xdouar,ydouar,hdouar,hdouarp
integer nnodedouar
double precision dtdouar,href
double precision dhminfluv,dhmaxfluv,dhmindiff,dhmaxdiff
double precision zerosion
double precision length_scale,velocity_scale,dt
integer baselevelx0,baselevelx1,baselevely0,baselevely1

integer(4) :: boundary_conditions, j, levelt, fastscape_nx, fastscape_ny, row
integer(4) :: fastscape_nstep

real(8) :: fastscape_dx, fastscape_dy
real(8) :: fastscape_dt
real(8), dimension(:,:), allocatable :: fastscape_z,  &
     fastscape_vx, fastscape_vy, fastscape_vz, fastscape_zprev, fluvial_erosion, &
     diffusion_erosion, fastscape_p
integer, dimension(:,:), allocatable :: fastscape_catchment

!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)

zerosion=params%zerosion
length_scale=params%length_scale
velocity_scale=params%velocity_scale
! only do half timestep in midpoint iteration
if (params%grid_converged) then
    dt=params%dt
else
    dt=params%dt/2.0d0
end if
baselevelx0=params%baselevelx0 ! ibc
baselevelx1=params%baselevelx1 ! ibc
baselevely0=params%baselevely0 ! ibc
baselevely1=params%baselevely1 ! ibc

if (is.ne.1) then

   if (surface%spread_surface_points.gt.0) then

      allocate (points(2,surface%nsurface),stat=err); if (err.ne.0) call stop_run ('Error alloc points in erode')
      allocate (pointsmin(2,surface%nsurface),stat=err) ; if (err.ne.0) call stop_run ('Error alloc pointsmin in erode')
      allocate (pointsmax(2,surface%nsurface),stat=err) ; if (err.ne.0) call stop_run ('Error alloc pointsmax in erode')
      allocate (ln_work(surface%nsurface),stat=err) ; if (err.ne.0) call stop_run ('Error alloc ln_work in erode')

      points=0.d0
      pointsmin=2.d0
      pointsmax=-1.d0
      ln_work=0
      do i=1,surface%nt
         do k=1,3
            kp=mod(k,3)+1
            ic=surface%icon(k,i)
            icp=surface%icon(kp,i)
            points(1,ic)=points(1,ic)+surface%x(icp)
            points(2,ic)=points(2,ic)+surface%y(icp)
            pointsmin(1,ic)=min(pointsmin(1,ic),surface%x(icp))
            pointsmax(1,ic)=max(pointsmax(1,ic),surface%x(icp))
            pointsmin(2,ic)=min(pointsmin(2,ic),surface%y(icp))
            pointsmax(2,ic)=max(pointsmax(2,ic),surface%y(icp))
            ln_work(ic)=ln_work(ic)+1
         enddo
      enddo

      do i=1,surface%nsurface
         if (surface%x(i).lt.1.d-10) ln_work(i)=0
         if (surface%x(i).gt.1.d0-1.d-10) ln_work(i)=0
         if (surface%y(i).lt.1.d-10) ln_work(i)=0
         if (surface%y(i).gt.1.d0-1.d-10) ln_work(i)=0
      enddo

      do i=1,surface%nsurface
         if (ln_work(i).ne.0) then
            if (surface%spread_surface_points.eq.1) then
               surface%x(i)=points(1,i)/ln_work(i)
               surface%y(i)=points(2,i)/ln_work(i)
            elseif (surface%spread_surface_points.eq.2) then
               surface%x(i)=(pointsmin(1,i)+pointsmax(1,i))/2.d0
               surface%y(i)=(pointsmin(2,i)+pointsmax(2,i))/2.d0
            elseif (surface%spread_surface_points.eq.3) then
               points(1,i)=points(1,i)/ln_work(i)
               points(2,i)=points(2,i)/ln_work(i)
            endif
         endif
      enddo

      if (surface%spread_surface_points.eq.3) then
      allocate (ppoints(2,surface%nsurface),stat=err) ; if (err.ne.0) call stop_run ('Error alloc ppoints in erode')
      allocate (wpoints(surface%nsurface),stat=err) ; if (err.ne.0) call stop_run ('Error alloc wpoints in erode')
      wpoints=0.d0
      do i=1,surface%nt
         do k=1,3
            kp=mod(k,3)+1
            ic=surface%icon(k,i)
            icp=surface%icon(kp,i)
            wpoints(ic)=wpoints(ic)+sqrt((surface%x(icp)-points(1,ic))**2+(surface%y(icp)-points(2,ic))**2)*0.1d0
         enddo
      enddo
      ppoints=0.d0
      do i=1,surface%nt
         do k=1,3
            kp=mod(k,3)+1
            ic=surface%icon(k,i)
            icp=surface%icon(kp,i)
            ppoints(1,ic)=ppoints(1,ic)+sqrt((surface%x(icp)-points(1,ic))**2+(surface%y(icp)-points(2,ic))**2)*0.1d0*surface%x(icp)
            ppoints(2,ic)=ppoints(2,ic)+sqrt((surface%x(icp)-points(1,ic))**2+(surface%y(icp)-points(2,ic))**2)*0.1d0*surface%y(icp)
         enddo
      enddo
      do i=1,surface%nsurface
         if (ln_work(i).ne.0) then
             surface%x(i)=ppoints(1,i)/wpoints(i)
             surface%y(i)=ppoints(2,i)/wpoints(i)
         endif
      enddo
      deallocate (ppoints,wpoints)
      endif

      deallocate (points)
      deallocate (pointsmin)
      deallocate (pointsmax)
      deallocate (ln_work)

   endif

   do i=1,surface%nsurface
      call octree_interpolate (olsf%octree,olsf%icon, &
                               surface%x(i),surface%y(i),surface%z(i),olsf%lsf,f)
      surface%z(i)=min(f,surface%z(i))
   enddo

   return

endif

! "is" is 1 from here: surface(is) == surface(1)
!
! first apply erosion to top surface


select case (params%erosion_type)
    case (NO_EROSION)
      ! Do nothing for now...
    case (FLAT_EROSION)
      ! Perfect erosrion
      do i=1,surface%nsurface
          if (surface%z(i) > zerosion) then
              surface%erosion(i) = (surface%z(i) - zerosion)/dt
              surface%z(i)=zerosion
          else
              surface%erosion(i) = 0.d0
          end if
      enddo
    case (CASCADE_EROSION)
      nnodedouar=surface%nsurface

      allocate (xdouar(nnodedouar))
      allocate (ydouar(nnodedouar))
      allocate (hdouar(nnodedouar))
      allocate (hdouarp(nnodedouar))

      xdouar=surface%x*length_scale*1.d3
      ydouar=surface%y*length_scale*1.d3

      call random_number (hdouar)
      hdouar=(hdouar+1.0)/2.d0
      hdouar=hdouar+(surface%z-zerosion)*length_scale*1.d3*params%vex
      href=min(minval(hdouar), 0.5d0)
      hdouar=hdouar-href
      dtdouar=dt*length_scale/velocity_scale*1.d6

      if (iproc.eq.0) then
          write (8,*) 'Cascade time scale ',dtdouar
          write (8,*) 'Topo in ',minval(hdouar),maxval(hdouar)
          write (8,*) 'X in ',minval(xdouar),maxval(xdouar)
          write (8,*) 'Y in ',minval(ydouar),maxval(ydouar)
      end if

      hdouarp=hdouar
      call cascade (xdouar,ydouar,hdouar,nnodedouar,dtdouar, &
                    dhminfluv,dhmaxfluv, &
                    dhmindiff,dhmaxdiff, &
                    fluvial_erosion,diffusion_erosion, &
                    baselevelx0,baselevelx1,baselevely0,baselevely1)

      if (iproc.eq.0) then
          write (8,*) 'Topo out ',minval(hdouar),maxval(hdouar)
          write (8,*) 'Delta Topo ',minval(hdouar-hdouarp),maxval(hdouar-hdouarp)
      end if

      do i=1,surface%nsurface
         surface%z(i)=(hdouar(i)+href)/1.d3/length_scale/params%vex+zerosion
         surface%erosion(i)=(hdouar(i)-hdouarp(i))/1.d3/length_scale/dt
      enddo

      deallocate (xdouar)
      deallocate (ydouar)
      deallocate (hdouar)
      deallocate (hdouarp)
    case (FASTSCAPE_EROSION)
      ! TODO
      !
      ! FastScapeSolo (z,nx,ny,dx,dy,dt,nstep,vx,vy,vz,kf,kd,p,m,n,ibc)
      ! z(nx,ny) double precision array of initial topography
      ! nx,ny are dimension of z
      ! dx and dy are x- and y-spacing
      ! dt is time step
      ! nstep is number of time steps
      ! vx(nx,ny),vy(nx,ny),vz(nx,ny) are the velocity fields
      ! kf(nx,ny) is the spatially variable fluvial erodibility constant
      ! kd(nx,ny) is the spatially variable hillslope diffusion constant
      ! p(nx,ny) is the spatially variable precipitation rate

      boundary_conditions = (1000 * baselevely0) + (100 * baselevelx1) + (10 * baselevely1) + baselevelx0  

      levelt = surface%levelt
      fastscape_nx = (2 ** levelt) + 2
      fastscape_ny = (2 ** levelt) + 2
      fastscape_dx = 1.d3 * length_scale / (fastscape_nx - 1)
      fastscape_dy = 1.d3 * length_scale / (fastscape_ny - 1)

      allocate(fastscape_z(fastscape_nx, fastscape_ny))
      allocate(fastscape_zprev(fastscape_nx, fastscape_ny))
      allocate(fastscape_vx(fastscape_nx, fastscape_ny))
      allocate(fastscape_vy(fastscape_nx, fastscape_ny))
      allocate(fastscape_vz(fastscape_nx, fastscape_ny))
      allocate(fastscape_p(fastscape_nx, fastscape_ny))
      allocate(fluvial_erosion(fastscape_nx, fastscape_ny))
      allocate(diffusion_erosion(fastscape_nx, fastscape_ny))
      allocate(fastscape_catchment(fastscape_nx, fastscape_ny))

      if (params%precipitation_type > 0) fastscape_p = surface%precipitation
      if (params%fluvial_erosion_type > 0) fluvial_erosion = surface%fluvial_erosion
      if (params%diffusion_erosion_type > 0) diffusion_erosion = surface%diffusion_erosion

      !-----------------------------------------------------------------------!
      ! Scale elevation, velocity and time from DOUAR to FastScape (m, a)
      !-----------------------------------------------------------------------!

      !seed z with random noise, increase elevation by [0.5,1] m
      call random_number (fastscape_z)
      fastscape_z=(fastscape_z+1.0)/2.d0

      do k = 1, fastscape_ny
        do j = 1, fastscape_nx
          row = (k - 1) * fastscape_ny
          fastscape_z(j, k) = fastscape_z(j, k) + (surface%z(j + row) - zerosion) *  &
                                 length_scale * 1.d3 * params%vex
          fastscape_vx(j, k) = surface%u(j + row) * velocity_scale * 1.0d-3
          fastscape_vy(j, k) = surface%v(j + row) * velocity_scale * 1.0d-3
          fastscape_vz(j, k) = surface%w(j + row) * velocity_scale * 1.0d-3
        enddo
      enddo

      call precipitation (params, fastscape_z, fastscape_p, fastscape_nx, fastscape_ny, fastscape_dx, fastscape_dy)
      call fluverosion (params, fastscape_z, fluvial_erosion, fastscape_nx, fastscape_ny, fastscape_dx, fastscape_dy)
      call differosion (params, fastscape_z, diffusion_erosion, fastscape_nx, fastscape_ny, fastscape_dx, fastscape_dy)

      !do not let baselevel rise above zerosion, e.g. by isostatic uplift)
      href=min(minval(fastscape_z), 0.5d0)
      fastscape_z=fastscape_z-href

      ! Fastscape time stepping
      fastscape_nstep = ceiling(dt / params%fastscape_dtmax)
      fastscape_dt = (dt / fastscape_nstep) * length_scale / velocity_scale * 1.d6

      if (iproc.eq.0) then
          write (8,*) 'Fastscape eroding in ',fastscape_nstep, &
                       'steps of ', fastscape_dt, ' years'
          write (8,*) 'Topo in ',minval(fastscape_z),maxval(fastscape_z)
          write (8,*) 'href is ',href
          write (8,*) 'dx and dy are', fastscape_dx, ' and ', fastscape_dy
          !call output_surf_nodevar (surface, is, 'surf_FS_eros', params%current_step, &
          !         2*params%current_griditer - 1, surface%w )
      end if

      fastscape_zprev = fastscape_z

      !if (iproc == 0 .and. params%grid_converged) then
      !    call output_fastscape_topo(fastscape_z, fastscape_nx, fastscape_ny,  &
      !         fastscape_dx, fastscape_dy, "before", params%current_step)
      !end if

      call FastScapeSolo (fastscape_z, fastscape_nx, fastscape_ny, &
                          fastscape_dx, fastscape_dy, &
                          fastscape_dt, fastscape_nstep, &
                          fastscape_vx, fastscape_vy, fastscape_vz,  &
                          fluvial_erosion, diffusion_erosion, fastscape_p, &
                          params, boundary_conditions, fastscape_catchment)

      if (iproc.eq.0) then
          write (8,*) 'Topo out ',minval(fastscape_z),maxval(fastscape_z)
          write (8,*) 'Delta Topo ',minval(fastscape_z-fastscape_zprev), &
                          maxval(fastscape_z-fastscape_zprev)
      end if

      !if (iproc == 0 .and. params%grid_converged) then
          !call output_fastscape_topo(fastscape_z, fastscape_nx, fastscape_ny,  &
               !fastscape_dx, fastscape_dy, "after", params%current_step)
      !end if

      do k = 1, fastscape_ny
        do j = 1, fastscape_nx
          row = (k - 1) * fastscape_ny
          surface%z(j + row) =  zerosion + (fastscape_z(j, k) + href) /  &
                               (1.0d3 * length_scale * params%vex)
          if (params%fastscape_minz.and.surface%z(j + row).lt.0.d0) then
              surface%z(j + row) = 0.d0
          endif
          surface%erosion(j + row) = (fastscape_z(j, k) - fastscape_zprev(j, k) - &
                                      fastscape_vz(j, k) * fastscape_dt * fastscape_nstep) /  &
                                       (1.0d3 * length_scale * dt)       
          surface%catchment(j + row) = fastscape_catchment(j, k)
        enddo
      enddo
      if (iproc.eq.0) then
          !call output_surf_nodevar (surface, is, 'surf_FS_eros', params%current_step, &
                   !2*params%current_griditer, surface%erosion )

      end if


      if (converge) then
         ! Store new calculated values back
         if (params%precipitation_type > 0) surface%precipitation = fastscape_p
         if (params%fluvial_erosion_type > 0) surface%fluvial_erosion = fluvial_erosion
         if (params%diffusion_erosion_type > 0) surface%diffusion_erosion = diffusion_erosion
      endif

      call mpi_barrier (mpi_comm_world, ierr)

      deallocate(fastscape_z)
      deallocate(fastscape_zprev)
      deallocate(fastscape_vx)
      deallocate(fastscape_vy)
      deallocate(fastscape_vz)
      deallocate(fastscape_p)
      deallocate(fluvial_erosion)
      deallocate(diffusion_erosion)
      deallocate(fastscape_catchment)

    case default
      print *, "erosion.f90: erosion type: unknown"
      stop
end select

if (params%sedimentation .AND. params%sedimentation_type == 0) then
    !Flat fill sedimentation
    do i=1,surface%nsurface
        if (surface%z(i) < zerosion) then
            surface%erosion(i) = surface%erosion(i) +  (surface%z(i) - zerosion)/dt
            surface%z(i)=zerosion
        end if
    enddo
end if

! nn_setup (calculates delaunay triangles and other info)
! This section calculates the interplated position of the free surface (surface 1) 
! for all nodes in olsf and stores the result in olsf%lsf. CAUTION: Contrary to its name, lsf here
! is not a level set function, but rather the value of surface(1)%z, for a given x and y coordinate,
! independent of z. 
! This value will be used for all lower surfaces to restrict the maximum z value.

ntmax=surface%nsurface*3
nhmax=surface%nsurface
npmax=surface%nsurface
nnpnmax=100
nmax=3*ntmax+npmax
nvmax=ntmax

allocate (points(2,surface%nsurface),stat=err); if (err.ne.0) call stop_run ('Error alloc points in erode')
allocate (field(surface%nsurface),stat=err)   ; if (err.ne.0) call stop_run ('Error alloc field in erode')
allocate (vertices(3,ntmax),stat=err)         ; if (err.ne.0) call stop_run ('Error alloc vertices in erode')
allocate (centres(3,ntmax),stat=err)          ; if (err.ne.0) call stop_run ('Error alloc centress in erode')
allocate (neighbour(3,ntmax),stat=err)        ; if (err.ne.0) call stop_run ('Error alloc neighbour in erode')
allocate (hulltriangles(nhmax),stat=err)      ; if (err.ne.0) call stop_run ('Error alloc hulltriangles in erode')
allocate (vis_tlist(nvmax),stat=err)          ; if (err.ne.0) call stop_run ('Error alloc vis_tlist in erode')
allocate (vis_elist(nvmax),stat=err)          ; if (err.ne.0) call stop_run ('Error alloc vis_elist in erode')
allocate (add_tlist(nvmax),stat=err)          ; if (err.ne.0) call stop_run ('Error alloc add_tlist in erode')
allocate (ln_work(surface%nsurface),stat=err) ; if (err.ne.0) call stop_run ('Error alloc ln_work in erode')
allocate (lt_work_l(ntmax),stat=err)            ; if (err.ne.0) call stop_run ('Error alloc lt_work_l in erode')
allocate (ln_work_l(surface%nsurface),stat=err) ; if (err.ne.0) call stop_run ('Error alloc ln_work_l in erode')

points(1,:)=surface%x
points(2,:)=surface%y
field(:)=surface%z
dmode=-2
nmode=0
clockwise=.true.
eps=1.d-20
eps=0.d0

call nn2d_setup (surface%nsurface,ntmax,nhmax,npmax,nnpnmax,nmax, &
                 points,dmode,nmode,clockwise,field,nt,vertices, &
                 centres,neighbour,nh,hulltriangles,nohalt_hull, &
                 loc,nnn,nnlist,ntrilist, &
                 eps,nvmax,vis_tlist,vis_elist,add_tlist, &
                 lt_work_l,ln_work_l)

allocate (work_d1(2,nnpnmax),stat=err) ; if (err.ne.0) call stop_run ('Error alloc work_d1 in erode')
allocate (work_d2(2,nnpnmax),stat=err) ; if (err.ne.0) call stop_run ('Error alloc work_d2 in erode')
allocate (work_d3(2,nnpnmax),stat=err) ; if (err.ne.0) call stop_run ('Error alloc work_d3 in erode')
allocate (work_d4(2,nnpnmax),stat=err) ; if (err.ne.0) call stop_run ('Error alloc work_d4 in erode')
allocate (work_d5(2,nnpnmax),stat=err) ; if (err.ne.0) call stop_run ('Error alloc work_d5 in erode')
allocate (work_d6(2,nnpnmax),stat=err) ; if (err.ne.0) call stop_run ('Error alloc work_d6 in erode')
allocate (work_d7(2,nnpnmax),stat=err) ; if (err.ne.0) call stop_run ('Error alloc work_d7 in erode')
allocate (work_i1(nnpnmax),stat=err)   ; if (err.ne.0) call stop_run ('Error alloc work_i1 in erode')
allocate (work_i2(2,nnpnmax),stat=err) ; if (err.ne.0) call stop_run ('Error alloc work_i2 in erode')
allocate (work_i3(nnpnmax),stat=err)   ; if (err.ne.0) call stop_run ('Error alloc work_i3 in erode')
allocate (work_r1(nnpnmax,2),stat=err) ; if (err.ne.0) call stop_run ('Error alloc work_r1 in erode')
allocate (work_r2(nnpnmax),stat=err)   ; if (err.ne.0) call stop_run ('Error alloc work_r2 in erode')
allocate (lsf(olsf%nnode),stat=err)    ; if (err.ne.0) call stop_run ('Error alloc lsf in erode')

nnext=.true.
int_method=3
loc=1

lsf=0.d0

do i=iproc+1,olsf%nnode,nproc
   x=olsf%x(i)
   y=olsf%y(i)
   call nn2D (x,y,points,vertices,neighbour,nnext,int_method,centres, &
              hulltriangles,nh,loc,field,nnpnmax,work_r1,work_r2, &
              work_d1,work_i1,work_i2,work_i3,work_d2,work_d3,work_d4, &
              work_d5,work_d6,work_d7,lt_work_l,ln_work_l,out,f,df,ddf,v)
   lsf(i)=f
enddo

olsf%lsf(1:olsf%nnode)=0.d0
call mpi_allreduce (lsf,olsf%lsf,olsf%nnode, &
                    mpi_double_precision,mpi_sum,mpi_comm_world,ierr)

! here spreads the points on the top surface to avoid collapse of the surface due to erosion
! for that it uses a barrycenter algorithm and a reinterpolation of the surface to avoid undesired diffusion effects

if (surface%spread_surface_points > 0 .AND. .NOT. surface%fastscape) then

allocate (pointsmin(2,surface%nsurface),stat=err) ; if (err.ne.0) call stop_run ('Error alloc pointsmin in erode')
allocate (pointsmax(2,surface%nsurface),stat=err) ; if (err.ne.0) call stop_run ('Error alloc pointsmax in erode')

points=0.d0
pointsmin=2.d0
pointsmax=-1.d0
ln_work=0
do i=1,surface%nt
  do k=1,3
  kp=mod(k,3)+1
  ic=surface%icon(k,i)
  icp=surface%icon(kp,i)
  points(1,ic)=points(1,ic)+surface%x(icp)
  points(2,ic)=points(2,ic)+surface%y(icp)
  pointsmin(1,ic)=min(pointsmin(1,ic),surface%x(icp))
  pointsmax(1,ic)=max(pointsmax(1,ic),surface%x(icp))
  pointsmin(2,ic)=min(pointsmin(2,ic),surface%y(icp))
  pointsmax(2,ic)=max(pointsmax(2,ic),surface%y(icp))
  ln_work(ic)=ln_work(ic)+1
  enddo
enddo
do i=1,surface%nsurface
if (surface%x(i).lt.1.d-10) ln_work(i)=0
if (surface%x(i).gt.1.d0-1.d-10) ln_work(i)=0
if (surface%y(i).lt.1.d-10) ln_work(i)=0
if (surface%y(i).gt.1.d0-1.d-10) ln_work(i)=0
enddo
do i=1,surface%nsurface
if (ln_work(i).ne.0) then
  if (surface%spread_surface_points.eq.1) then
  surface%x(i)=points(1,i)/ln_work(i)
  surface%y(i)=points(2,i)/ln_work(i)
  else
  surface%x(i)=(pointsmin(1,i)+pointsmax(1,i))/2.d0
  surface%y(i)=(pointsmin(2,i)+pointsmax(2,i))/2.d0
  endif
endif
enddo

deallocate (pointsmin,pointsmax)

endif

deallocate (points)
deallocate (field)
deallocate (vertices)
deallocate (centres)
deallocate (hulltriangles)
deallocate (vis_tlist)
deallocate (vis_elist)
deallocate (add_tlist)
deallocate (lt_work_l)
deallocate (ln_work)
deallocate (ln_work_l)
deallocate (work_d1)
deallocate (work_d2)
deallocate (work_d3)
deallocate (work_d4)
deallocate (work_d5)
deallocate (work_d6)
deallocate (work_d7)
deallocate (work_r1)
deallocate (work_r2)
deallocate (work_i1)
deallocate (work_i2)
deallocate (work_i3)
deallocate (lsf)

return
end subroutine erosion
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|


subroutine output_fastscape_topo(fastscape_z, fastscape_nx, fastscape_ny, fastscape_dx, fastscape_dy, prefix, current_step)
  implicit none

  real(8), intent(in), dimension(:,:) :: fastscape_z
  real(8), intent(in) :: fastscape_dx, fastscape_dy

  integer(4), intent(in) :: fastscape_nx, fastscape_ny, current_step
  integer(4) :: i, j
  integer(4), parameter :: file_unit = 90

  character(*), intent(in) :: prefix
  character(5) :: step_as_str

  write(step_as_str, "(I5.5)") current_step

  open(unit = file_unit, file = prefix // "_fastscape_topo_" // step_as_str // ".vtk")

  write(file_unit, "(A)") "# vtk DataFile Version 3.0"
  write(file_unit, "(A)") "Surface"
  write(file_unit, "(A)") "ASCII"
  write(file_unit, "(A)") "DATASET STRUCTURED_POINTS"
  write(file_unit, *) "DIMENSIONS ", fastscape_nx, fastscape_ny, 1
  write(file_unit, *) "ORIGIN ", 0.0, 0.0, 0.0
  write(file_unit, *) "SPACING ", fastscape_dx, fastscape_dy, 1.0
  write(file_unit, *) "POINT_DATA ", fastscape_nx * fastscape_ny
  write(file_unit, *) "SCALARS H float 1"
  write(file_unit, *) "LOOKUP_TABLE default"

  do j = 1, fastscape_ny
    do i = 1, fastscape_nx
      write(file_unit, *) fastscape_z(i, j)
    enddo
  enddo

  close(file_unit)

end subroutine output_fastscape_topo

!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|


end module m_erosion
