module m_initialize_temperature
contains
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              ||===\\                                                         | 
!              ||    \\                                                        |
!              ||     ||   //==\\   ||  ||   //==||  ||/==\\                   |
!              ||     ||  ||    ||  ||  ||  ||   ||  ||    ||                  |
!              ||    //   ||    ||  ||  ||  ||   ||  ||                        |
!              ||===//     \\==//    \\==\\  \\==\\  ||                        |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              INITIALIZE_TEMPERATURE    Nov. 2006                             |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

subroutine initialize_temperature (ov,surface,mat,params)

!------------------------------------------------------------------------------|
!(((((((((((((((( Purpose of the routine  ))))))))))))))))))))))))))))))))))))))
!------------------------------------------------------------------------------|
! this routine initializes the remperature field to some basic
! conductive equilibrium base on sp01 (baseline) of the surfaces
! temperature should be normalized between 0 and 1
! ov is the velocity octree containing the velocity/temperature solution
! this routine needs to be improved...

!------------------------------------------------------------------------------|
!((((((((((((((((  declaration of the subroutine arguments  ))))))))))))))))))))
!------------------------------------------------------------------------------|
      
use definitions
use m_toolbox
use m_octree

implicit none

include 'mpif.h'

type (octreev) :: ov
type (parameters) :: params
type (sheet) :: surface(params%ns)
type (material) :: mat(0:params%nmat)


!------------------------------------------------------------------------------|
!(((((((((((((((( declaration of the subroutine internal variables )))))))))))))
!------------------------------------------------------------------------------|

type (octreev) :: readtov

logical isostasyout,isobcout,nestout,compactionout,ptracking,calc_hpressure

integer iproc, nproc, ierr, err
integer nface, nlsf, np, vermajor,verminor,verstat,verrev
integer :: i,j,k,matnum,nstep,noctree
integer :: steptop,stepbottom

double precision :: distance,dist, current_time
double precision :: diff,heat,dz,zstep,delz

integer :: matindex(0:500)
double precision :: tgrad(0:500)
double precision, dimension (:), allocatable :: temp

!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)

select case (params%tempinit)
case (0)
!inear gradient form bottom to ztemp
    do i=1,ov%nnode
        ov%temp(i)=max((params%ztemp-ov%z(i))/params%ztemp,0.d0)
    enddo

case (1)
!steady state temperatures bases on surface baselines (sp01) and material heat productions
    allocate (temp(0:500), stat=err)
    nstep = 500
    dz = params%ztemp*params%vex/nstep

    !find surface closest above position and set material index to the corresponding 
    !material number

    matindex=0
    matnum=0
    do i=0,nstep
        matnum = 0 
        distance = -2.d0
        do j=1,params%ns
            dist = i*dz-surface(j)%sp01
            if (dist>distance .and. dist<=0.d0) then
                distance = dist
                matnum = surface(j)%material
            end if
        end do
        matindex(i) = matnum
    end do

    temp=0.d0
    tgrad=0.d0

    !first run of temperature init: no temp flux at bottom
    !initialize arrays
    temp(0)=1.d0
    tgrad(0)=0.d0

    do i=1,nstep
        !calculate diffusivity and heat as mean of bottom and top position
        diff = (mat(matindex(i))%diffusivity+mat(matindex(i-1))%diffusivity)*0.5d0
        heat = (mat(matindex(i))%heat+mat(matindex(i-1))%heat)*0.5d0

        tgrad(i) = tgrad(i-1) + heat*dz/diff
        temp(i) = temp(i-1) - (tgrad(i)+tgrad(i-1))*0.5*dz
    end do


    !determine bottom flux as missing temperature
    tgrad(0) = temp(nstep)/(params%ztemp*params%vex)
    temp(0) = 1.d0

    !Redo with bottom heat flux
    do i=1,nstep
        !calculate diffusivity and heat as mean of bottom and top position
        diff = (mat(matindex(i))%diffusivity+mat(matindex(i-1))%diffusivity)*0.5d0
        heat = (mat(matindex(i))%heat+mat(matindex(i-1))%heat)*0.5d0

        tgrad(i) = tgrad(i-1) + heat*dz/diff
        temp(i) = temp(i-1) - (tgrad(i)+tgrad(i-1))*0.5*dz

        !write (*,*) temp(i), tgrad(i)
    end do

    !set temperature by interpolating between closest two points from temperature calculation
    do i=1,ov%nnode
        zstep = ov%z(i)*params%vex/dz
        steptop = ceiling(zstep)
        stepbottom = floor(zstep)
        if (steptop > nstep) then
            ov%temp(i) = 0.d0
        else if (steptop == stepbottom) then
            ov%temp(i) = temp(stepbottom)
        else 
            delz = mod(zstep,1.d0)
            ov%temp(i) = delz*temp(stepbottom)+(1.d0-delz)*temp(steptop)
        end if
        ov%temp(i) = max(0.d0,min(1.d0,ov%temp(i)))
    end do
    deallocate (temp)

case (-1)
!read temperatures from octree of defined file

    !read input file
    readtov%noctree=params%noctreemax
    allocate (readtov%octree(readtov%noctree),stat=err)
    if (err.ne.0) call stop_run ('Error alloc readtov%octree in initialize_temperature')

    ! reads the temperature octree from a file
    open (25,file=trim(params%tempinitfile),status='old',form='unformatted')
    ! Read version number
    read (25) vermajor,verminor,verstat,verrev
    if (vermajor/=params%vermajor .or. verminor/=params%verminor .or. verstat/=params%verstat) then
      if (iproc==0) then
        write(*,'(a,i1,a,i1,a,i1,a,i1,a,i1,a,i1,a)') 'DOUAR-WSMP v', &
            params%vermajor,'.',params%verminor,'.',params%verstat, &
            ' cannot use output from v',vermajor,'.',verminor,'.',verstat, ' to initialize temperatures'
        write(*,'(a)') 'because the output formats are not compatible. To initialize with this output file'
        write(*,'(a,i1,a,i1,a,i1,a)') 'please use DOUAR-WSMP v', &
            params%vermajor,'.',params%verminor,'.',params%verstat,'.'
      endif
      call stop_run ('Version conflict in reading tempinit file')
    endif
    ! Read optional output flags
    read (25) isostasyout,isobcout,nestout,compactionout,ptracking,calc_hpressure
    read (25) !params%velocity_scale, params%length_scale, params%density_scale, params%tempscale, params%vex
    read (25) readtov%octree(3),readtov%nnode,readtov%nleaves,nface,nlsf,np,current_time
    noctree = readtov%octree(3)
    allocate (readtov%x(readtov%nnode),stat=err)
    if (err.ne.0) call stop_run ('Error alloc readtov%x in initialize_temperature')
    allocate (readtov%y(readtov%nnode),stat=err)
    if (err.ne.0) call stop_run ('Error alloc readtov%y in initialize_temperature')
    allocate (readtov%z(readtov%nnode),stat=err)
    if (err.ne.0) call stop_run ('Error alloc readtov%z in initialize_temperature')
    allocate (readtov%icon(8,readtov%nleaves),stat=err)
    if (err.ne.0) call stop_run ('Error alloc readtov%icon in initialize_temperature')
    allocate (readtov%temp(readtov%nnode),stat=err)
    if (err.ne.0) call stop_run ('Error alloc readtov%temp in initialize_temperature')


    read (25) readtov%x
    read (25) readtov%y
    read (25) readtov%z
    read (25) !readtov%unode
    read (25) !readtov%vnode
    read (25) !readtov%wnode
    read (25) !readtov%lsf
    read (25) readtov%temp
    read (25) !readtov%temporary_nodal_pressure
    read (25) !readtov%strain
    read (25) !readtov%kfix
    read (25) !readtov%kfixt

    if (isostasyout) read (25) !readtov%wnodeiso
    if (compactionout) read (25) !readtov%wnodecompact
    read (25) readtov%icon
    read (25) !rreadtov%pressure
    read (25) !rreadtov%spressure
    if (calc_hpressure) read (25) !readtov%hpressure
    read (25) !readtov%crit
    read (25) !readtov%e2d
    read (25) !readtov%eviscosity
    read (25) !readtov%is_plastic
    read (25) !readtov%dilatr
    read (25) !readtov%matnum
    read (25) !readtov%yield_ratio
    read (25) !readtov%frict_angle
    read (25) !readtov%whole_leaf_in_fluid

    if (compactionout) read (25) !readtov%compaciton_density

    read (25) readtov%octree(1:noctree)
    close (25)

    !interpolate onto octree
    allocate (temp(ov%nnode),stat=err)
    if (err.ne.0) call stop_run ('Error alloc temp in initialize_temperature')
    temp=0.d0

    do k=1+iproc,ov%nnode,nproc
        call octree_interpolate_octree (readtov%octree, readtov%icon, &
                                 readtov%x, readtov%y, readtov%z, &
                                 ov%x(k), ov%y(k), ov%z(k), readtov%temp, temp(k))
    enddo

    ov%temp=0.d0
    call mpi_allreduce (temp, ov%temp, ov%nnode, mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    deallocate (temp)

    !deallocate readtov
    deallocate (readtov%octree)
    deallocate (readtov%x)
    deallocate (readtov%y)
    deallocate (readtov%z)
    deallocate (readtov%icon)
    deallocate (readtov%temp)

case (3)
    do i=1,ov%nnode
        ov%temp(i)=0.d0
    enddo

case default
    call stop_run ('Please choose correct tempinit option')

end select

return
end subroutine initialize_temperature 
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|


end module m_initialize_temperature

