module m_refine_surface
contains
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              ||===\\                                                         | 
!              ||    \\                                                        |
!              ||     ||   //==\\   ||  ||   //==||  ||/==\\                   |
!              ||     ||  ||    ||  ||  ||  ||   ||  ||    ||                  |
!              ||    //   ||    ||  ||  ||  ||   ||  ||                        |
!              ||===//     \\==//    \\==\\  \\==\\  ||                        |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              REFINE_SURFACE    May 2008                                      |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
      
subroutine refine_surface(params,surface,threadinfo,nadd,nrem,is,&
           istep,ref_count) 

!------------------------------------------------------------------------------|
!(((((((((((((((( Purpose of the routine  ))))))))))))))))))))))))))))))))))))))
!------------------------------------------------------------------------------|
! This routine builds an edge array between a set of particles on a surface.
! It uses the delaunay triangulation and then steps through the triangles
! to build the edge information.
! ed is the computed edge array
! nedge is the number of edges
! nedgepernode,nodeedgenumber and nodenodenumber contain the list of edges 
! that start from each node; for a given node i
! their number is nedgepernode(i), the edge number in the list of edges is
! nodeedgenumber(j,i) and the node at the end of the edge is
! nodenodenumber(j,i) for j=1,nedgepernode(i)

! this routine builds the boolean 'refine' array of length nedge (number of edges 
! in the triangulation connecting the particles) that contains the list of 
! edges to be refined.  

! surface is the sheet/surface to be refined
! it will contain the new number of particles
! ed is the computed edge array
! refine is the integer array determining the edges to be refined
! nedge is the number of edges
! nadd is the number of edges to be changed/split
! nedgepernode, nodenodenumber,nodeedgenumber are arrays containing edge info
! nnmax is the maximum number of nn

!------------------------------------------------------------------------------|
!((((((((((((((((  declaration of the subroutine arguments  ))))))))))))))))))))
!------------------------------------------------------------------------------|
      
use definitions
!use mpi
use threads
use m_heap
use m_toolbox
use m_check_delaunay
use m_remove_multiple_points
use m_compute_normals
use m_qsort
use m_remove_bad

implicit none

include 'mpif.h'

type (parameters) params
type (sheet) surface 
type (thread) threadinfo
integer nadd,nrem
integer is,istep
integer ref_count

!------------------------------------------------------------------------------|
!(((((((((((((((( declaration of the subroutine internal variables )))))))))))))
!------------------------------------------------------------------------------|

type (sheet) surface0 
type (edge), dimension(:), allocatable :: ed
type (edge), dimension(:), allocatable :: edswap

integer :: naddp, nnmax
integer :: ie,j,jedge,kp,kpp,k,inode,inodep,inodepp,iadd,irem
integer :: ntriangle, nedge, nedgen, nsurfacen
integer :: err,ierr,iproc,nproc,i,iedge,i1,i2
integer :: nnodeint,nelemint,nedgeint,nelemmax,nnodemax,nedgemax
integer :: rancalls, maxrancalls, nremoved, nneighb, nrem1
integer,dimension(:)  ,allocatable :: nedgepernode
integer,dimension(:)  ,allocatable :: refine_list,remove_list,remove_list1
integer,dimension(:,:),allocatable :: nodenodenumber,nodeedgenumber,allneighbours
integer,dimension(:,:),allocatable :: imemswap, iconswap

double precision dist1,distmax,distmin,eps
double precision xm,ym,zm,xyzn,ran
double precision,dimension(:),allocatable::remaddlist
double precision,dimension(:,:),allocatable::memswap

logical,dimension(:),allocatable::refine,refinep,remove,removep

character*72  :: shift


!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)
         
shift=' '
eps=1.d-8


!------------------------
! build edge information 
!------------------------

nsurfacen=surface%nsurface
ntriangle=surface%nt
nedge=nsurfacen+ntriangle-1
if (surface%closed.eq.1) nedge=nedge-1
nnmax=12 ! nnmax is maximum number of neighbours in triangulation (should be around 6 on average)

allocate (allneighbours(-1:13,surface%nsurface))
allneighbours = 0

do i=1,surface%nt
    !for each triangle, check for identical nodes
    checktri: do j=1,3
      do k = 0,1
        !process both neighbours of node i in this triangle
        inode=surface%icon(j,i)
        kp=mod(j+k,3)+1
        inodep=surface%icon(kp,i)

        if (inode /= inodep) then
        !add to neighbour_list
            if (allneighbours(0,inode) == 0) then
                allneighbours(0,inode) = 1
                allneighbours(1,inode) = inodep
            else
                !only add if not already in neighbour list
                nneighb = allneighbours(0,inode) 
                if (count( allneighbours(1:nneighb,inode) == inodep) == 0) then 
                    allneighbours(0,inode) = nneighb+1
                    allneighbours(nneighb+1,inode) = inodep
                end if
            end if
        else
            ! triangle has identical nodes as corners
            if (ref_count > 0) then
                call remove_bad_triangles (surface, surface0, 0, is, params, ref_count, nremoved)
                ref_count = -1 * ref_count
                return
            else
                call stop_run ("Bad triangles not solved by remove_bad_triangles")
            end if
        end if
      end do
    end do checktri
end do

!check number of neighbours for all nodes
do i = 1,surface%nsurface
    if (surface%convexhull(i) == 0) then
        if (allneighbours(0,i) < 3) then
            if (ref_count > 0) then
                call remove_bad_triangles (surface, surface0, 0, is, params, ref_count, nremoved)
                ref_count = -1 * ref_count
                return
            else
                call stop_run ("Bad triangles not solved by remove_bad_triangles")
            end if
        elseif (allneighbours(0,i) == 3) then
            !test maximum distance
            do j=1,3
                distmax = 0.d0
                i2 = allneighbours(j,i)
                dist1=dist(surface%x(i),surface%x(i2),surface%y(i),surface%y(i2),         &
                    surface%z(i),surface%z(i2))
                distmax = max(distmax,dist1)
            end do
            !add to remove list
            if (distmax < 2.d0**(-surface%levelt-2)) allneighbours(-1,i) = 1
        end if
    end if
end do

nrem1 = sum(allneighbours(-1,:))
allocate(remove_list1(1:nrem1))
irem = 0
do i = 1,surface%nsurface
    if (allneighbours(-1,i) == 1) then
        irem = irem+1
        remove_list1(irem) = i
    end if
end do
deallocate (allneighbours)

if (iproc.eq.0 .and. params%debug>=1 .and. irem>0) then
    write(*,'(a,i2,a,i8,a)') shift//'S.',is,':', irem,' ptcls with 3 neighbours found in refine_surface'
end if


allocate (ed(nedge),stat=threadinfo%err)
if (params%debug.gt.1) call heap (threadinfo,'ed','refine_surface',size(ed),'int',+1)
allocate (nedgepernode(nsurfacen),stat=threadinfo%err)
if (params%debug.gt.1) call heap (threadinfo,'nedgepernode','refine_surface',size(nedgepernode),'int',+1)
allocate (nodenodenumber(nnmax,nsurfacen),stat=threadinfo%err)
if (params%debug.gt.1) call heap (threadinfo,'nodenodenumber','refine_surface',size(nodenodenumber),'int',+1)
allocate (nodeedgenumber(nnmax,nsurfacen),stat=threadinfo%err)
if (params%debug.gt.1) call heap (threadinfo,'nodeedgenumber','refine_surface',size(nodeedgenumber),'int',+1)

ed%t1=0
ed%t2=0
nedgepernode=0
iedge=0

do ie=1,ntriangle !loop over all triangles
  do k=1,3 !loop over three points in triangle
    !get node numbers of the points that define this triangle
    inode=surface%icon(k,ie)
    kp=mod(k,3)+1
    kpp=mod(k+1,3)+1
    inodep=surface%icon(kp,ie)
    inodepp=surface%icon(kpp,ie)

    !get edge information of right hand triangle
    do j=1,nedgepernode(inodep)
      if (nodenodenumber(j,inodep).eq.inode) then
        jedge=nodeedgenumber(j,inodep)
        ed(jedge)%m2=inodepp
        ed(jedge)%t2=ie
        goto 111
      endif
    enddo

    !check edge number and connectivity
    iedge=iedge+1
    if (iedge.gt.nedge) then
       if (ref_count > 0) then
           call remove_bad_triangles (surface, surface0, 0, is, params, ref_count, nremoved)
           ref_count = -1 * ref_count
           return
       else
           if (iproc==0) then
                write(8,*) 'surface: nsurface=',surface%nsurface,'ntriangle=',surface%nt
                write(8,*) 'internal: nsurfacen=',nsurfacen,'ntriangle=',ntriangle
                write(8,*) 'nedge=',nedge,' iedge=',iedge
                call output_surf(surface,is,"crash_edge_refine_surf",istep,ref_count) 
           end if
           call stop_run ('too many edges in refine_surface')
       end if
   end if
    nedgepernode(inode)=nedgepernode(inode)+1
    if (nedgepernode(inode).gt.nnmax) then
        if (iproc.eq.0) then
            write (8,*) 'Node ',inode,' has ',nedgepernode(inode),' neighbors'
            write (*,*) 'Node ',inode,' has ',nedgepernode(inode),' neighbors'
        end if
        if (ref_count > 0) then
            call remove_bad_triangles (surface, surface0, 0, is, params, ref_count, nremoved)
            ref_count = -1 * ref_count
            return
        else
            if (iproc == 0) call output_surf(surface,is,"crash_edge_refine_surf",istep,ref_count) 
            call stop_run ('too many neighbours in refine_surface')
        end if
    end if

    !get edge information of left hand triangle and start end end points
    nodenodenumber(nedgepernode(inode),inode)=inodep
    nodeedgenumber(nedgepernode(inode),inode)=iedge
    ed(iedge)%n1=inode
    ed(iedge)%n2=inodep
    ed(iedge)%m1=inodepp
    ed(iedge)%t1=ie

    111 continue
  enddo
enddo

if (iedge/=nedge) then 
    if (ref_count > 0) then
        call remove_bad_triangles (surface, surface0, 0, is, params, ref_count, nremoved)
        ref_count = -1 * ref_count
        return
    else
        if (iproc.eq.0) then
            write (8,*) "iedge /= nedge in refine_surface"
            write (8,*) 'iedge=',iedge,'nedge=',nedge
            write (*,*) "iedge /= nedge in refine_surface"
            call output_surf(surface,is,"crash_edge_refine_surf",istep,ref_count) 
        end if
        call stop_run ('iedge/=nedge in refine_surface')
    end if
end if

!------------------------------------
! find edges that need to be refined
!------------------------------------

distmax=1.d0/(2.d0**surface%levelt+1)*sqrt(2.d0)*surface%stretch   !opla
distmin=1.d0/(2.d0**(surface%levelt+4)+1)*sqrt(2.d0)

allocate (refine(nedge),stat=err)
allocate (refinep(nedge),stat=err)
refine=.false.
refinep=.false.

allocate (remove(nedge),stat=err)
allocate (removep(nedge),stat=err)
remove=.false.
removep=.false.

rancalls = 0
do iedge=1+iproc,nedge,nproc
   i1=ed(iedge)%n1
   i2=ed(iedge)%n2

   if ((count(remove_list1==i1)>0) .OR. (count(remove_list1==i2)>0)) cycle

   dist1=dist(surface%x(i1),surface%x(i2),surface%y(i1),surface%y(i2),         &
             surface%z(i1),surface%z(i2))
   ! Add noise to point injection
   call random_number(ran)
   rancalls = rancalls+1
   refinep(iedge)=(dist1.gt.distmax*(ran+2.5d0)/3.d0)

   !force removal when too close for convex hull edges
   if (params%remove_surf_pts .OR. &
     (surface%convexhull(i1)>0 .AND. surface%convexhull(i2)>0)) then
       removep(iedge)=(dist1.gt.eps .and. dist1.le.distmin)
   end if

   !make sure no conflicts - remove before refine
   refinep(iedge) = (refinep(iedge) .AND. (.NOT. removep(iedge)))
end do

!even out calls to random_number
call mpi_allreduce (rancalls,maxrancalls,1,mpi_integer,mpi_max,mpi_comm_world,ierr)
do i=rancalls+1,maxrancalls
    call random_number(ran)
end do

call mpi_allreduce (refinep,refine,nedge,mpi_logical,mpi_lor,mpi_comm_world,ierr)
call mpi_allreduce (removep,remove,nedge,mpi_logical,mpi_lor,mpi_comm_world,ierr)

!-----------------------
! compute nadd and naddp
!-----------------------

!since the convex hull is not refined, can we assume that naddp is always 0
!and remove it from calculation?
! LMK 2015-10-06: ... but what if we allow convex hull refinement?

nrem=count(remove)+nrem1
allocate(remove_list(nrem))
remove_list = 0
remove_list(1:nrem1) = remove_list1(1:nrem1)
deallocate (remove_list1)
nremoved = 0

do iedge=1,nedge
  if (remove(iedge)) then
    i1 = ed(iedge)%n1
    i2 = ed(iedge)%n2
    !prefer to remove node with smaller convex hull number
    if ((surface%convexhull(i1) <= surface%convexhull(i2)) .AND. &
        (.NOT. btest(surface%convexhull(i1),0)) )  then
      irem = irem + 1
      remove_list(irem)=i1
    !make sure not to remove corner
    elseif (.NOT. btest(surface%convexhull(i2),0)) then
      irem = irem+1
      remove_list(irem)=i2
    !should not happen: two corner points are connected
    elseif (btest(surface%convexhull(i1),0) .AND. btest(surface%convexhull(i2),0)) then
      if (iproc==0) then
          write (*,*) 'WARNING: found two corner nodes connected in refine_surface'
          write (*,'(a,I9,a,3F10.7,a,I3)') 'n1 = ',i1, 'at ', & 
              surface%x(i1), surface%y(i1),surface%z(i1), &
              ', convexhull = ', surface%convexhull(i1)
          write (*,'(a,I9,a,3F10.7,a,I3)') 'n2 = ',i2, 'at ', & 
              surface%x(i2), surface%y(i2),surface%z(i2), &
              ', convexhull = ', surface%convexhull(i2)

          write (8,*) 'WARNING: found two corner nodes connected in refine_surface'
          write (8,'(a,I9,a,3F10.7,a,I3)') 'n1 = ',i1, 'at ', & 
              surface%x(i1), surface%y(i1), surface%z(i1), &
              ', convexhull = ', surface%convexhull(i1)
          write (8,'(a,I9,a,3F10.7,a,I3)') 'n2 = ',i2, 'at ', & 
              surface%x(i2), surface%y(i2), surface%z(i2), &
              ', convexhull = ', surface%convexhull(i2)
      end if
    else
      !should never get here
      if (iproc==0) then
          write (*,*) 'Convex hull calculation is wrong'
      end if
    end if
  endif
enddo

nadd=count(refine)
allocate(refine_list(nadd))
iadd=0
naddp=0

do iedge=1,nedge
  if (refine(iedge)) then
    !check if n1 or n2 are in remove list, if so, do not refine
    i1 = ed(iedge)%n1
    i2 = ed(iedge)%n2
    if ((count(remove_list==i1) > 0) .OR. (count(remove_list==i2) > 0)) then
        !do not refine edge
        if (iproc == 0) then
            write (*,*) "WARNING: edge marked for refinement, but one edge node will be removed"
            write (8,*) "WARNING: edge marked for refinement, but one edge node will be removed"
        end if
    else
        iadd=iadd+1 
        if (ed(iedge)%t1.eq.0 .or. ed(iedge)%t2.eq.0) naddp=naddp+1

        refine_list(iadd)=iedge
    end if
  endif
end do
nadd = iadd

if (params%debug > 1) then
    allocate (remaddlist(surface%nsurface))
    remaddlist = 0.d0

    do i=1,irem
        remaddlist(remove_list(i)) = -3.d0
    end do
    do i=1,iadd
        iedge = refine_list(i)
        remaddlist(ed(iedge)%n1) = 1.d0
        remaddlist(ed(iedge)%n2) = 1.d0
    end do

    call output_surf_nodevar(surface,is,'refine',istep,(ref_count-1)*2+1,remaddlist)

    deallocate (remaddlist)
end if

deallocate (refine)
deallocate (refinep)
deallocate (remove)
deallocate (removep)

if (iproc.eq.0 .and. params%debug>=1) then
    write(*,'(a,i2,a,i8,a)') shift//'S.',is,':', nadd,' added ptcls in refine_surface'
end if

!-----------------------------------
! resizing ed (allocate/deallocate)
!-----------------------------------

nedgen=(surface%nsurface+nadd)+(surface%nt+(nadd-naddp)*2+naddp)-1
if (surface%closed.eq.1) nedgen=nedgen-1

allocate (edswap(nedgen),stat=err) ; if (err.ne.0) call stop_run ('Error alloc edswap in main')
edswap(1:nedge)=ed(1:nedge)
deallocate (ed)
allocate (ed(nedgen),stat=err) ; if (err.ne.0) call stop_run ('Error alloc edswap in main')
ed(1:nedge)=edswap(1:nedge)
deallocate (edswap)

!----------------------------------
! prepare memory for the refinment
!----------------------------------

nnodemax=surface%nsurface+nadd
nelemmax=surface%nt+(nadd-naddp)*2+naddp
nedgemax=nnodemax+nelemmax-1
if (surface%closed.eq.1) nedgemax=nedgemax-1

if (nedgemax.ne.nedgen) call stop_run ('error in counting number of edges in refine_surface')

allocate (memswap(surface%nsurface,6),stat=err) ; if (err.ne.0) call stop_run('Error alloc memswap in refine_surface')
allocate (imemswap(surface%nsurface,1),stat=err) ; if (err.ne.0) call stop_run('Error alloc imemswap in refine_surface')
allocate (iconswap(3,surface%nt),stat=err) ; if (err.ne.0) call stop_run('Error alloc iconswap in refine_surface')
      
memswap(:,1)=surface%x
memswap(:,2)=surface%y
memswap(:,3)=surface%z
memswap(:,4)=surface%xn
memswap(:,5)=surface%yn
memswap(:,6)=surface%zn
imemswap(:,1)=surface%convexhull
iconswap(1:3,1:surface%nt)=surface%icon(1:3,1:surface%nt)

deallocate (surface%x)
deallocate (surface%y)
deallocate (surface%z)
deallocate (surface%xn)
deallocate (surface%yn)
deallocate (surface%zn)
deallocate (surface%icon)
deallocate (surface%convexhull)

allocate (surface%x(nnodemax), stat=err) ; if (err.ne.0) call stop_run ('Error alloc x in refine_surface')
allocate (surface%y(nnodemax), stat=err) ; if (err.ne.0) call stop_run ('Error alloc y in refine_surface')
allocate (surface%z(nnodemax), stat=err) ; if (err.ne.0) call stop_run ('Error alloc z in refine_surface')
allocate (surface%xn(nnodemax),stat=err) ; if (err.ne.0) call stop_run ('Error alloc xn in refine_surface')
allocate (surface%yn(nnodemax),stat=err) ; if (err.ne.0) call stop_run ('Error alloc yn in refine_surface')
allocate (surface%zn(nnodemax),stat=err) ; if (err.ne.0) call stop_run ('Error alloc zn in refine_surface')
allocate (surface%convexhull(nnodemax), stat=err) ; if (err.ne.0) call stop_run ('Error alloc convexhull in refine_surface')
allocate (surface%icon(3,nelemmax),stat=err) ; if (err.ne.0) call stop_run ('Error alloc icon in refine_surface')
      
surface%x(1:surface%nsurface)=memswap(:,1)
surface%y(1:surface%nsurface)=memswap(:,2)
surface%z(1:surface%nsurface)=memswap(:,3)
surface%xn(1:surface%nsurface)=memswap(:,4)
surface%yn(1:surface%nsurface)=memswap(:,5)
surface%zn(1:surface%nsurface)=memswap(:,6)
surface%convexhull(1:surface%nsurface)=imemswap(:,1)
surface%icon(1:3,1:surface%nt)=iconswap(1:3,1:surface%nt)

deallocate (memswap)
deallocate (imemswap)
deallocate (iconswap)

!---------------------
! updates the surface
!---------------------

nnodeint=surface%nsurface
nelemint=surface%nt
nedgeint=nedge

do i=1,nadd
   iedge=refine_list(i)
   i1=ed(iedge)%n1
   i2=ed(iedge)%n2
   if (i1 == i2) then
       if (iproc == 0) then
         write(*,*) '** *****************************************************************'
         write(*,*) 'Nodes for edge ',i,' are identical before refinement!'
         write(*,*) 'The edge node is ', i2, 'at :'
         write(*,'(a,f12.8,a,f12.8,a,f12.8)') 'x:',surface%x(i1),' y: ',surface%y(i1),' z: ',surface%z(i1)
         write(*,*) 'Bad triangles will be removed'
         write(*,*) '*******************************************************************'
       end if
    
       if (ref_count > 0) then
           call remove_bad_triangles (surface, surface0, 0, is, params, ref_count, nremoved)
           ref_count = -1 * ref_count
           return
       else
           call stop_run ("Bad triangles not solved by remove_bad_triangles")
       end if
   endif
   xm=(surface%x(i2)+surface%x(i1))*0.5d0
   ym=(surface%y(i2)+surface%y(i1))*0.5d0
   zm=(surface%z(i2)+surface%z(i1))*0.5d0
   nnodeint=nnodeint+1
   surface%x(nnodeint)=xm
   surface%y(nnodeint)=ym
   surface%z(nnodeint)=zm
   xm=(surface%xn(i1)+surface%xn(i2))*0.5d0
   ym=(surface%yn(i1)+surface%yn(i2))*0.5d0
   zm=(surface%zn(i1)+surface%zn(i2))*0.5d0
   xyzn=sqrt(xm**2+ym**2+zm**2)
   surface%xn(nnodeint)=xm/xyzn
   surface%yn(nnodeint)=ym/xyzn
   surface%zn(nnodeint)=zm/xyzn

   ! Check whether the new surface point is part of the convex hull
   ! Convex hull flagging of the points is:
   !              8 (y=1)
   !        25.-------.13
   !          |       |
   ! 16 (x=0) |       | 4 (x=1)
   !          |       |
   !        19.-------.7
   !              2 (y=0)
   !
   
   !bitwise integer and results is common edge, also works with corners
   !If two corners are connected (should not happen), new node is not of convex hull
   
   surface%convexhull(nnodeint) = IAND(surface%convexhull(i1), surface%convexhull(i2))
   if (btest(surface%convexhull(nnodeint),0)) then
       surface%convexhull(nnodeint)=0
   endif
   call update_icon (surface%icon,nelemmax,ed,nedgemax,iedge,nnodeint,nelemint,nedgeint)
enddo

do i=1,size(ed%n1)
   i1=ed(i)%n1
   i2=ed(i)%n2
   if (i1 == i2) then
       if (iproc == 0) then
         write(*,*) '*******************************************************************'
         write(*,*) 'Nodes for edge ',i,' are identical after refinement!'
         write(*,*) 'The edge node is ', i2, 'at :'
         write(*,'(a,f12.8,a,f12.8,a,f12.8)') 'x:',surface%x(i1),' y: ',surface%y(i1),' z: ',surface%z(i1)
         write(*,*) 'Bad triangles will be removed'
         write(*,*) '*******************************************************************'
       end if
    
       if (ref_count > 0) then
           surface%nsurface=nnodemax
           surface%nt=nelemmax
           call remove_bad_triangles (surface, surface0, 0, is, params, ref_count, nremoved)
           ref_count = -1 * ref_count
           return
       else
           call stop_run ("Bad triangles not solved by remove_bad_triangles")
       end if
   endif
enddo

surface%nsurface=nnodemax
surface%nt=nelemmax

if (irem>0) then
    call remove_multiple_points(remove_list(1:irem), irem, surface, 0, surface0, is, params, nremoved)
endif

if (iproc.eq.0 .and. params%debug>=1 .and. irem>0) then
    write(*,'(a,i2,a,i8,a)') shift//'S.',is,':', nremoved,' removed ptcls in refine_surface'
end if

if (.not.params%normaladvect) then
   call compute_normals (surface%nsurface,surface%x,surface%y,surface%z, &
                         surface%nt,surface%icon,surface%xn,surface%yn,surface%zn)
end if

if (nnodeint.ne.nnodemax) call stop_run ('Error counting nodes in refine_surface')
if (nelemint.ne.nelemmax) call stop_run ('Error counting elements in refine_surface')

if (nremoved.gt.0) then
  nnodemax=surface%nsurface
  nelemmax=surface%nt
  nedgemax=nnodemax+nelemmax-1
endif

surface%nsurface=nnodemax
surface%nt=nelemmax
nedge=nedgemax
!adjust nrem to actual number removed
nrem = nremoved

deallocate(refine_list)
deallocate(remove_list)

if (params%debug.gt.1) call heap (threadinfo,'ed','refine_surface',size(ed),'int',-1)
deallocate (ed)         
if (params%debug.gt.1) call heap (threadinfo,'nedgepernode','refine_surface',size(nedgepernode),'int',-1)
deallocate (nedgepernode)
if (params%debug.gt.1) call heap (threadinfo,'nodenodenumber','refine_surface',size(nodenodenumber),'int',-1)
deallocate (nodenodenumber)
if (params%debug.gt.1) call heap (threadinfo,'nodeedgenumber','refine_surface',size(nodeedgenumber),'int',-1)
deallocate (nodeedgenumber)

ref_count = abs(ref_count)
if (params%debug > 1) call output_surf(surface,is,'refine',istep,(ref_count-1)*2+2) 

end subroutine refine_surface

!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
  
subroutine update_icon (icon,nelemmax,ed,nedgemax,iedge,nnodeint,nelemint,nedgeint)
  
use definitions
  
implicit none
  
integer icon(3,nelemmax),nelemmax,nedge,iedge,nnodeint,nelemint
integer jedge,nedgemax,nedgeint
type(edge) ed(nedgemax)
integer k,k12,k21
integer n1,n2,m1,m2,t1,t2,t1n,t2n
  
nedge=nedgeint

n1=ed(iedge)%n1
n2=ed(iedge)%n2
m1=ed(iedge)%m1
m2=ed(iedge)%m2
t1=ed(iedge)%t1
t2=ed(iedge)%t2

if (t1.ne.0) then
   do k=1,3
      if (icon(k,t1).eq.n1) k12=k
   enddo
   icon(k12,t1)=nnodeint
   nelemint=nelemint+1
   t1n=nelemint
   icon(1,nelemint)=m1
   icon(2,nelemint)=n1
   icon(3,nelemint)=nnodeint
endif

if (t2.ne.0) then
   do k=1,3
      if (icon(k,t2).eq.n2) k21=k
   enddo
   icon(k21,t2)=nnodeint
   nelemint=nelemint+1
   t2n=nelemint
   icon(1,nelemint)=m2
   icon(2,nelemint)=n2
   icon(3,nelemint)=nnodeint
endif

ed(iedge)%n2=nnodeint
if (t1.ne.0) then
   ed(iedge)%t1=t1n
else
   ed(iedge)%t1=0
endif

if (t1.ne.0) then
   nedgeint=nedgeint+1
   ed(nedgeint)%n1=m1
   ed(nedgeint)%n2=nnodeint
   ed(nedgeint)%m1=n2
   ed(nedgeint)%m2=n1
   ed(nedgeint)%t1=t1
   ed(nedgeint)%t2=t1n
endif

nedgeint=nedgeint+1
ed(nedgeint)%n1=n2
ed(nedgeint)%n2=nnodeint

if (t1.ne.0) then
   ed(nedgeint)%m2=m1
   ed(nedgeint)%t2=t1
else
   ed(nedgeint)%t2=0
endif

if (t2.ne.0) then
   ed(nedgeint)%m1=m2
   ed(nedgeint)%t1=t2n
else
   ed(nedgeint)%t1=0
endif

if (t2.ne.0) then
   nedgeint=nedgeint+1
   ed(nedgeint)%n1=m2
   ed(nedgeint)%n2=nnodeint
   ed(nedgeint)%m1=n1
   ed(nedgeint)%m2=n2
   ed(nedgeint)%t1=t2
   ed(nedgeint)%t2=t2n
endif

do jedge=1,nedge
   if (t1.ne.0) then
      if (ed(jedge)%t1.eq.t1) then
         if (ed(jedge)%n1.eq.m1) then
            ed(jedge)%m1=nnodeint
            ed(jedge)%t1=t1n
         endif
         if (ed(jedge)%n1.eq.n2) ed(jedge)%m1=nnodeint
      endif
      if (ed(jedge)%t2.eq.t1) then
         if (ed(jedge)%n1.eq.n1) then
            ed(jedge)%m2=nnodeint
            ed(jedge)%t2=t1n
         endif
         if (ed(jedge)%n1.eq.m1) ed(jedge)%m2=nnodeint
      endif
   endif
   if (t2.ne.0) then
      if (ed(jedge)%t1.eq.t2) then
         if (ed(jedge)%n1.eq.m2) then
            ed(jedge)%m1=nnodeint
            ed(jedge)%t1=t2n
         endif
         if (ed(jedge)%n1.eq.n1) ed(jedge)%m1=nnodeint
      endif
      if (ed(jedge)%t2.eq.t2) then
         if (ed(jedge)%n1.eq.n2) then
            ed(jedge)%m2=nnodeint
            ed(jedge)%t2=t2n
         endif
         if (ed(jedge)%n1.eq.m2) ed(jedge)%m2=nnodeint
      endif
   endif
enddo

return
end subroutine update_icon

!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
end module m_refine_surface

